/* Simulation of the Hawaii chip. (Includes the ADC device, assumes my wiring configuration) 		*
 * Reads in a simulation logfile (from pb_parse), outputs (hopefully) a .cam format file of ADC data 	*/

/* Copyright (C) Richard Neill 2012-2013, <hawaiisim at REMOVE.ME.richardneill.org>. This program is Free Software. You can   
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <tiffio.h>
#include <stdarg.h>
#include "hawaiisim.h"

/* Virtual Hawaii Array */
int col[QUADS];							/* The horizontal (fast) registers. Define the column (x-coord), controlled by pixel/lsync. Use -1 for "before the first pixel clock" */
int row[QUADS];							/* The vertical (slow) registers. Define the row (y-coord), controlled by line/fsync.  Use -100 for "uninitialised". */
signed int pixels[QUADS][NUM_ROWS][NUM_COLS];			/* The actual pixel values in the array. Controlled by light, time and noise. Array is [quad][row][col]  i.e. slow, fast */
unsigned int pixels_read_count[QUADS][NUM_ROWS][NUM_COLS];	/* The number of times each pixel has been read since it was reset */
unsigned long long pixels_reset_ns[QUADS][NUM_ROWS][NUM_COLS];	/* The reset timestamps for each pixel */
unsigned long long pixels_lastupdated_ns[QUADS][NUM_ROWS][NUM_COLS]; /* Last updated timestamp. Used to do illumination with lazy_evaluation (about 1M times faster algorithm!) */
int read_enabled;						/* Are pixel reads enabled? (all quads) */

/* Other (global) variables */
FILE *output_fh; char outfile[FILENAME_LEN]; 			/* output file  */
unsigned int write_buf[QUADS];					/* write buffer for output. 4 quadrants. */
int resetmode, exposemode, adcmode;				/* modes for reset/adc/exposure */
typedef uint32_t sample_t;					/* 4 bytes per sample for the tiff. */
sample_t pixels_tiff[NUM_ROWS*2][NUM_COLS*2];			/* Tiff pixels - note that this is indexed differently to the hawaii pixels */
unsigned int debug=0, conterr=0, conterr_count=0, quiet=0, test=0; /* Flags: debug on? Continue on error? */
unsigned int inbus, inbus_prev;					/* input data bus via "pulseblaster": current step and previous one. */
unsigned int inbus_lo, inbus_hi, inbus_fe, inbus_re, inbus_ee; 	/* input as low, high, falling-edge, rising-edge, either-edges */
unsigned long long re_at_ns[PB_BITS], fe_at_ns[PB_BITS], ee_at_ns[PB_BITS]; /* when each bit last changed (rising edge, falling edge, either edge) */
unsigned long long length_ns, elapsed_ns=0;			/* instruction length, elapsed time since start */
unsigned int reset_level, reset_tconst_ns, gencam_reset_ns;	/* Parameters */
unsigned long long gencam_exposure_ns;
double glow_ens, light_ens, tiff_ens, background_ens, hawaii_noise, adc_noise, mains_fhz, mains_hum_level, mains_phase=0;
int line_num, instr_count=0, read_count=0, reset_count=0, sqw_count=0, clip_count=0; /* misc */

/* Function prototypes */
void print_help(), fatal_error2(), initialise_chip(), reset_lines(), lazy_illuminate_pixels(), lazy_illuminate_array(), read_enable(), read_disable(), read_quads();
void clock_pixel(int q), clock_line(int q), lsync(), frsync(), simulate_step(), gen_cam();
int dither(double ditherand), parse_sourceline (char buffer[BUFFER_LEN], int line_num);
unsigned int rand_n (unsigned int n), shot_noise(double lambda);
double rand_u(), thermal_noise (unsigned int sigma);

/* Helpful macros */
#define eprintf(...)    fprintf(stderr, __VA_ARGS__)                            /* Error printf: send to stderr  */
#define deprintf(...)   if(debug){ fprintf(stderr, __VA_ARGS__);}               /* Debugging error printf: send to stderr  */
#define feprintf(...)   fprintf(stderr, __VA_ARGS__); exit (EXIT_FAILURE)       /* Fatal error printf: send to stderr and exit */


/* Documentation */
void print_help(){
	fprintf (stdout,
		"INTRODUCTION:\n"
		"	Hawaiisim simulates the Rockwell Hawaii_25 chip and the NI4462 ADC converter hardware,\n"
		"	using a PulseBlaster replay log (.%s) file. The simulation file contains the sequence of\n"
		"	control-data timings written to the virtual hawaii. The resulting output data is a series\n"
		"	of analog readings taken from the virtual hawaii chip, clocked by the %s file, and exposed\n"
		"	to light/glow/noise at a certain rate. Hawaiisim also carefully verifies that all clock\n"
		"	pulses are legal (eg no counter overflows, min-lengths satisfied, no read during reset, no\n"
		"	adc with read disabled etc). The particular wiring of the Hawaii chip is #define'd to\n"
		"	correspond to the physical wiring in the dewar. Copyright %s, %s, %s.\n"
		"\n"
		"	This is useful for several purposes:\n"
		"	  - Validation of pulseblaster programs.\n"
		"	  - Checking which pixels are read/reset (address, count, sequence).\n"
		"	  - Checking whether the program is mains-synchronous.\n"
		"	  - Simulating image generation for analysis.\n"
		"\n"
		"NOTE:\n"
		"	The real hardware uses an NI4462 ADC, which begins acquisition of many samples when triggered by ADC.\n"
		"	It's more helpful to simulate a FAKEADC which does exactly ONE conversion, on every falling-edge of\n"
		"	the FAKEADC bit. Multiple-reads must be explicit in the replaylog.\n"
		"\n"
		"EXAMPLE:\n"
		"	hawaiisim -i frame-reset-read.pbsim -o tmp.cam -e glow -L 0.5 -r noise; \n"
		"	cam2tiff tmp.cam tmp.tiff; display tmp.tiff \n"
		"	This replays a full-frame pattern, simulates the exposure, and draws the image.\n"
		"	See: /usr/local/share/doc/hawaiisim/ for example files.\n"
		"\n",
		EXTN_PBSIM, EXTN_PBSIM, AUTHOR, YEAR, LICENSE);

	fprintf (stdout,
		"USAGE:  hawaiisim  [options] -i LOGFILE.%s -o OUTPUT.%s\n"
		"\n"
		"OPTIONS:\n"
		"       -h, --help              Show help\n"
		"\n"
		"       -i  INFILE.%s        Input File. (required, except with -t)\n"
		"       -o  OUTFILE.%s         Output File (otherwise, infile, renamed with .%s)  Or %s.\n"
		"\n"
		"       -t                      Test mode. Ignore infile; output a full-array image in .%s format.\n"
		"                               Reset for %.2f us, expose for %.2f ms, do ADC for %.2f ms.\n"
		"\n"
		"       -d                      Debugging on (very verbose).\n"
		"       -c                      Continue on error, don't exit immediately. (Eventually, exit 1).\n"
		"       -q                      Quieter: don't print pixel addresses for reads/resets (also faster).\n"
		"       -x                      Allow overwrite of existing output file (also, don't enforce file-extensions).\n"
		"\n"
		"       -a  ADC_MODE            ADC simulation model: ideal | noise | hum\n"
		"             ideal   [default]   ADC reads perfectly.\n"
		"             noise               Each ADC read has thermal noise, mean %.2f electrons.\n"
		"             hum     [best]      Thermal noise (%.2f e) and %.2f Hz mains hum (%.2f electrons).\n"
		"\n"
		"       -r  RESET_MODE            Reset model: zero | const | noise | expon\n"
		"             zero    [default]   Reset is perfect, to exactly zero.\n"
		"             const               Reset is perfect, to level %d electrons.\n"
		"             noise               Reset to constant %d with thermal kTC noise %.2f electrons.\n"
		"             expon   [best]      Cells decay exponentially towards level %d, with\n"
		"                                 time-constant %.2f us and thermal noise %.2f electrons.\n"
		"\n"
		"       -e  EXPOSURE_MODE       Exposure model: debug | uniform | glow | debug:file | linear:file | image:file\n"
		"            debug                Pixels are set to own address (quad:row:col) + noise in lsb.\n"
		"            uniform  [default]   Uniform, linear exposure at %.2f electrons/us. No noise.\n"
		"            glow                 Linear (%.2f e/us), with glow and background (%.2f + %.2f e/us).\n"
		"            debug:FILE           Pixels are set to the exact value from FILE (no integration).\n"
		"            linear:FILE          Linear exposure, pixels from FILE. White pixels are %.2f e/us.\n"
		"            image:FILE  [best]   Exposure from file (%.2f e/us) with glow, background (%.2f + %.2f e/us).\n"
		"                               To create a correctly formatted image FILE.%s, use gencamtest/cam2tiff.\n"
		"\n"
		"CONFIGURATION OPTIONS:         (changes to the default values)\n"
		"       -A  ADC_R_NOISE         Set ADC read (resistor) noise std-deviation. Default: %.2f electrons.\n"
		"       -H  HAWAII_C_NOISE      Set Hawaii thermal (capacitor-reset) noise. Default: %.2f electrons.\n"
		"       -R  RESET_LEVEL         Set Hawaii reset level. Default %d electrons.\n"
		"       -T  RESET_TIME_CONST    Set Hawaii reset time-constant (for -r expon). Default: %.2f us.\n"
		"       -G  GLOW_LEVEL          Set Hawaii source-follower glow level (shot-noise). Default %.2f e/us.\n"
		"       -B  BACKGROUND_LEVEL    Set Hawaii background 'glow' level (shot-noise). Default %.2f e/us.\n"
		"       -L  LIGHT_LEVEL         Set illumination level (shot-noise mean). Default %.2f e/us.\n"
		"       -I  IMAGE_LEVEL         Set image level for a white pixel. (shot-noise). Default %.2f e/us.\n"
		"       -E  TEST_EXPOSURE_LEN   Set exposure length in test (-t) mode. Default %.2f us.\n"
		"       -F  TEST_RESET_LEN      Set reset length in test (-t) mode. Default %.2f us.\n"
		"       -V  MAINS_HUM_LEVEL     Set mains hum amplitude. Default: %.2f electrons.\n"
		"       -W  MAINS_HUM_FREQ      Set mains hum frequency. Default: %.2f Hz.\n"
		"\n",
		EXTN_PBSIM, EXTN_DAT, EXTN_PBSIM, EXTN_DAT, EXTN_DAT, DEV_NULL, EXTN_CAM,
		(double)gencam_reset_ns/NS_US, (double)gencam_exposure_ns/NS_MS, (double)(MIN_INTERVAL_FAKEADC_NS*NUM_ROWS*NUM_COLS)/NS_MS,
		adc_noise, adc_noise, mains_fhz, mains_hum_level, reset_level, reset_level, hawaii_noise, reset_level, (double)reset_tconst_ns/NS_US, hawaii_noise,
		light_ens*NS_US, light_ens*NS_US, glow_ens*NS_US, background_ens*NS_US, tiff_ens*NS_US, tiff_ens*NS_US, glow_ens*NS_US, background_ens*NS_US, EXTN_TIFF,
		adc_noise, hawaii_noise, reset_level, (double)reset_tconst_ns/NS_US, glow_ens*NS_US, background_ens*NS_US, light_ens*NS_US, tiff_ens*NS_US, (double)gencam_exposure_ns/NS_US, (double)gencam_reset_ns/NS_US, mains_hum_level, mains_fhz
		);

	fprintf (stdout,
		"PHYSICS MODEL:\n"
		"	Each array pixel is considered to be a uint_24, the same as the ADC output. We use integer\n"
		"	values, and work in electrons (rather than photons) for convenience. The ADC is configured\n"
		"	such that 1 ADC_level = 1 electron = 1 unit. Adding light increases the value. Conversion\n"
		"	of float -> int is dithered, so summing exposures over short intervals integrates correctly.\n"
		"	Pixel voltages are int_32, so can temporarily exceed the range (eg reset + negative noise).\n"
		"\n"
		"	RESET causes the pixels in that row to decay exponentially (with a very short time-constant)\n"
		"	towards the level at which they are biased. Thermal capacitor-reset noise 'HAWAII_C_NOISE'.\n"
		"	is added. (The physics model can be changed with -r, and the parameters by -R/-T; the best\n"
		"	'real-world' model is '-r expon').\n"
		"\n"
		"	ADC reads obtain the pixel value, with thermal (resistor) noise 'ADC_R_NOISE'. Multiple\n"
		"	reads reduce the noise by SQRT(n). The mains-hum noise has amplitude 'MAINS_HUM_LEVEL', and\n"
		"	frequency 'MAINS_HUM_FREQ'. The ADC is constrained to the range [0,0x%x]; values outside this\n"
		"	will clip. (The physics model can be changed with -a, and the parameter by -A; the best\n"
		"	'real-world' model is '-a hum'). Also, the 'FakeADC is an ideal, virtual ADC that makes a single\n"
		"	conversion exactly when triggered, the real-world ADC has all sorts of 'quirks' including a\n"
		"	variable delay for triggering, and convolution between adjacent (multiplexed) samples.\n"
		"\n"
		"	ILLUMINATION/Exposure to light exposes the array to a certain number of photons (electrons)\n"
		"	proportional to the brightness and the time. These arrive according to the shot-noise model.\n"
		"	The array may be illuminated uniformly, or the illumination pattern specified by a TIFF.\n"
		"	Glow noise and Background noise (shot-noise) are added; glow depends on the output-follower\n"
		"	There is NO thermal noise added by illumination: an isolated capacitor cannot spontaneously\n"
		"	alter its charge; also the total of such noise would depend on the mathematical exposure \n"
		"	interval, dt ! (The physics model can be changed with -e, and the parameters by -G/-L/-I;\n"
		"	the best 'real-world' model is '-e glow' or '-e tiff:file').\n"
		"\n"
		"	Noise is modelled as thermal noise or shot noise or mains hum.\n"
		"	 * THERMAL NOISE is Gaussian, mean=0, and is used for ADC_R_NOISE and HAWAII_C_NOISE.\n"
		"	   Capacitance-reset noise is incurred by Reset; thermal resistance noise by ADC conversion.\n"
		"	 * SHOT NOISE is Poisson, always > 0, and is used to model incoming light for exposure: i.e.\n"
		"	   GLOW_LEVEL (zero, unless the output-follower is read-enabled), BACKGROUND_LEVEL, and \n"
		"	   LIGHT_LEVEL / IMAGE_LEVEL. This smooths-out over long (or sum of short) exposures.\n"
		"	   For Resistor and Capacitor noise models: http://en.wikipedia.org/wiki/Johnson–Nyquist_noise .\n"
		"	 * MAINS HUM is ideal: MAINS_HUM_LEVEL * cos(2*PI*MAINS_HUM_FREQ). Harmonics and spikes not\n"
		"	   included. Ideally small, given tri-axial cable and battery-powered earth-loop avoidance.\n"
		"\n"
		"	IMPERFECTIONS in the model:\n"
		"	 * In the model, pixels reset to zero, and fill towards 0x%x. In the real device, reset\n"
		"	   charges the FPA and photons discharge each cell towards zero.\n"
		"	 * The ADC range is defined as [0,0x%x]; it maps exactly to full-scale without headroom.\n"
		"	 * Glow is uniform, but the Hawaii's Glow originates from the source followers in the corners.\n"
		"\n",
		 MAX_24BIT, MAX_24BIT, MAX_24BIT);

	fprintf (stdout,
		"CIRCUIT WIRING CONFIGURATION:\n"
		"	This is the 'wiring' between the Hawaii array and the individual bits of the PulseBlaster,\n"
		"	#define'd to match the circuit in the dewar. Only the lowest %d bits are used. Some lines\n"
		"	are commoned across quadrants: (READ,RESETB,LSYNC,FSYNC).\n"
		"\n"
		"	  BIT		FUNCTION	EXPLANATION\n"
		"	  ---   	---------	-----------\n"
		"	   %2d		PIXEL   0	These clock the horizontal (x,column,fast) register on\n"
		"	   %2d		PIXEL   1	either edge; even-numbered(**) columns are selected after a\n"
		"	   %2d		PIXEL   2	positive edge; odd colums after a negative edge.\n"
		"	   %2d		PIXEL   3	[Must be low when LSync is low.]\n"
		"\n"
		"	   %2d		LINE    0	These clock the vertical (y,row,slow) register on either\n"
		"	   %2d		LINE    1	edge; even-numbered rows(**) are selected after a positive\n"
		"	   %2d		LINE    2	edge; odd rows after a negative edge.\n"
		"	   %2d		LINE    3	[Must be low when LSync is low.]\n"
		"\n"
		"	   %2d		READ   all	Active high enables output follower; low minimises glow.\n"
		"	   %2d		RESETB all	Active low to reset the currently selected rows.\n"
		"	   %2d		LSYNC  all	Active low to reset the column register on all quadrants.\n"
		"	   %2d		FSYNC  all      Active low to reset the row register on all quadrants.\n"
		"\n"
		"	   %2d		ADC    		Trigger the real NI4462 ADC on a falling edge.\n"
		"	   %2d		FAKEADC		Trigger the FAKEADC to take one sample on a falling edge.\n"
		"\n"
		"	Pulse length requirements are configured as:\n"
		"	 * Any line:    minimum pulse (either phase):    %d ns\n"
		"	 * Reset line:  minimum pulse: 		         %d ns\n"
		"	 * ADC line:    minimum pulse:		         %d ns\n"
		"	 * Fake_ADC:    minimum pulse:		         %d ns\n"
		"	 * Fake_ADC:    setup/hold times:	         %d/%d ns\n"
		"	 * Fake_ADC:    minimum interval for retrigger:  %d ns\n"
		"\n"
		"       (**) Note: the Hawaii docs count from 1..512; we count 0..511.\n"
		"\n",
		PB_BITS, PIXEL_0_BIT, PIXEL_1_BIT, PIXEL_2_BIT, PIXEL_3_BIT, LINE_0_BIT, LINE_1_BIT, LINE_2_BIT, LINE_3_BIT, READ_BIT, RESETB_BIT, LSYNC_BIT, FSYNC_BIT, ADC_BIT, FAKEADC_BIT,
		MIN_PULSE_POSN_NS, MIN_PULSE_RESET_NS, MIN_PULSE_ADC_NS, MIN_PULSE_FAKEADC_NS, MIN_INTERVAL_FAKEADC_NS, MIN_SETUP_READ_FAKEADC_NS, MIN_HOLD_READ_FAKEADC_NS);

	fprintf(stdout,
		"ORDERING OF PIXELS AND QUADRANTS:\n"
		"	The Quadrant/Pixel/Raster orientation in the Hawaii documentation; also in `cam2tiff -h`.\n"
		"	Quadrants 0,1,2,3 are BR,BL,TL,TR. Quadrant origin is bottom left: x moves right; y up.\n"
		"	 * ROW    = y (or c) = slow-direction, controlled by LINE / FSYNC. (RESETB is whole row).\n"
		"	 * COLUMN = x (or r) = fast-direction, controlled by PIXEL / LSYNC\n"
		"	Row is more significant than column, hence co-ordinates are listed as (y,x).\n"
		"\n"
		"	The input TIFF file (used by -e tiff:file) is a standard %d x %d %d bit greyscale image.\n"
		"	Similarly to cam2tiff, it is 'the right way up', (even though X11's conventions have the\n"
		"	origin top-left, with y downwards.). The LSByte of the TIFF is ignored; the 3 MSB control\n"
		"	the output (such that 0x%x or white is worth %.2f e/us.)\n"
		"\n"
		"	Each ADC read results in a stream of 4 words (intel le_32 byte-order) for Q0,Q1,Q2,Q3.\n"
		"	If the entire array is raster-scanned once (eg -t), the result will be a valid .%s file.\n"
		"	Otherwise, the output will be a stream of 16-byte groups: use the .%s extension for the output.\n"
		"\n"
		"	The Hawaii chip has %d rows/columns per quadrant, addressed as 0...%d. After reset (LSync/FSync),\n"
		"	before the first clock-edge (Pixel/Line), no pixel is connected to the output. Clocking beyond \n"
		"	the end is technically ok, but more than one off the end (%d) is considered an error. Likwise, \n"
		"	reading from a disconnected row/column (off the end, or not at the beginning) triggers an error.\n"
		"\n",
		NUM_COLS*2, NUM_ROWS*2, 8*(int)sizeof(sample_t), MAX_24BIT, tiff_ens*NS_US, EXTN_CAM, EXTN_DAT, NUM_ROWS, NUM_ROWS-1, NUM_ROWS);

	fprintf(stdout,
		"STREAMS:\n"
		"	* The simulated ADC data is written to outfile. This is in the .cam format (use cam2dat to convert).\n"
		"	* The pixel addresses of reads/resets are sent to STDOUT (unless -q).\n"
		"	* Errors (and debugging info, if -d) go to STDERR.\n"
		"\n"
		"EXIT STATUS:\n"
		"	%d on success, %d on error.\n"
		"\n",
		EXIT_SUCCESS, EXIT_FAILURE);
}


/* If fatal error occurred, print message, then delete the dest file. This prevents accidentally using it for subsequent processing. */
/* Note that fatal errors are, in fact, non-fatal if -c is used. */
#define fatal_error(...)   fprintf(stderr, __VA_ARGS__); fatal_error2();
void fatal_error2(){
	if (instr_count == 0){	/* If pc is 0, then don't die. Some error conditions occur on first instr as a consequence of random PB start condition. */
		eprintf ("A 'fatal' error at instruction 0 can sometimes happen because of the randomised start-state of the chip. Ignoring...\n");
		conterr_count++;
	}else if (conterr){
		eprintf ("A fatal error occurred. Trying to continue anyway (invoked with -c). (Error occurred at pc: %d, source_line %d).\n", instr_count, line_num);
		conterr_count++;
	}else{
		if (strcmp (outfile, DEV_NULL)){   /* don't try to delete /dev/null */
			unlink (outfile);
			feprintf("A fatal error occurred. Output file %s has been deleted to prevent accidentally using it. (Error occurred at pc: %d, source_line %d).\n", outfile, instr_count, line_num);
		}else{
			feprintf("A fatal error occurred. (Error occurred at pc: %d, source_line %d).\n", instr_count, line_num);
		}
	}
}

/* Return a linearly-distributed random unsigned integer in the range 0...n. (inclusive)    (Assumes RAND_MAX is 2147483647) */
unsigned int rand_n (unsigned int n){			/* This isn't an especially good scaling, but it will do, especially for small numbers */
	if (n > RAND_MAX){				/* RAND_MAX is the largest value to fit in the positive half of a signed int. */
		fatal_error ("Sorry, too large a random number request. You asked for %d, but RAND_MAX is %d\n", n, RAND_MAX);
	}
	n += 1;
	return (rand() % n);
}

/* Return a uniformly-distributed double, in the range [0,1] */
double rand_u(){
	return ((double)rand() / RAND_MAX);  /* Non-trivial time taken here, when looped over multiple pixels and multiple time-intervals */
}

/* Return an unsigned integer, from shot noise (Poisson) distribution, parameterised by mean = lambda */
unsigned int shot_noise(double lambda){
	int k;
	long double L, p;	/* Too large? Use normal approximation instead. (experimentally, 10000 is about the limit for correct calculation or sane speed) */
	if (lambda > 25){	/* However, "the approximation is good when lambda > 10" , so use 25, and benefit from the faster calculation speed */
		L = round ( thermal_noise (sqrt(lambda)) );   	/* Normal (0, SQRT(lambda) ).  */
		L = L + lambda;					/* shift */
		if (L < 0){
			return (0);
		}else{
			return (int)(L);
		}
	}else{			/* Poisson. Use Algorithm by Knuth. */
		L = expl(-lambda);    k = 0;    p = 1;
		do{
			k++;
			p = p * rand_u();	/* uniform [0,1] */
		} while (p > L);
		return (k - 1);
	}
}

/* Return a double, from thermal noise (Gaussian) distribution, parameterised by mean = 0, variance = sigma */
double thermal_noise (unsigned int sigma){	/* Use Central Limit theorem to approximate: sum 12 U[0,1] deviates, then subtract 6. (and scale). */
	int i;
	double z;
	z = 0;
	for (i=0; i<12; i++){
		z += rand_u();   /* uniform [0,1] */
	}
	z -= 6;
	return (z * sigma);
}

/* Convert float to int. Rounding up/down with probability based on the fractional part. */
int dither(double ditherand){
	int ans;
	ans = (int)ditherand;
	if (ditherand >= 0){
		if ( ((ditherand - floor(ditherand)) * RAND_MAX) > rand() ){
			ans++;
		}
	}else{
		if ( ((ceil(ditherand) - ditherand) * RAND_MAX) > rand() ){
			ans--;
		}
	}
	return (ans);
}

/* Does a given pixel exist? Many arrays only defined for the sensible values, don't segfault. */
int array_key_exists (int r0, int c0, int r1, int c1, int r2, int c2, int r3, int c3){
	return (  (r0 >= 0) && (r0 < NUM_ROWS) &&  (r1 >= 0) && (r1 < NUM_ROWS) &&  (r2 >= 0) && (r2 < NUM_ROWS) &&  (r3 >= 0) && (r3 < NUM_ROWS) &&
	(c0 >= 0) && (c3 < NUM_COLS) &&  (c1 >= 0) && (c3 < NUM_COLS) &&  (c2 >= 0) && (c3 < NUM_COLS) &&  (c3 >= 0) && (c3 < NUM_COLS) );
}

/* Set the arrays and data to random states, as if just powered on. Useful as a bug check! */
void initialise_chip(){
	int r,c,q,bit;

	inbus_prev = rand_n(MAX_24BIT) & USED; 	 /* Previous state of pulseblaster outputs "before" we began the program. */
	read_enabled = inbus_prev & READ;	 /* Is it better to put in random data (the power-on state is unknown, or invalid data? */
	for (q=0; q< QUADS; q++){		
		col[q] = RC_UNINIT; /* rand_n (NUM_COLS);  was random, now changed to invalid */
		row[q] = RC_UNINIT; /*rand_n (NUM_COLS);   */	
	}
	for (q=0 ; q < QUADS; q++){
		for (r=0; r<NUM_ROWS; r++){
			for (c=0; c<NUM_COLS; c++){
				pixels[q][r][c] = rand_n (MAX_24BIT - (MAX_24BIT >> 6)); /* random (positive, non-saturated) data in pixel */
				pixels_read_count[q][r][c] = 0;			/* hasn't been read yet */
				pixels_lastupdated_ns[q][r][c] =0; 		/* updated value = 0 (for lazy evaluation) */
				pixels_reset_ns[q][r][c] = 0x8fffffffffffffffLL;  /* hasn't been reset. Use half the max of unsigned long long, to cause obvious garbage */
			}
		}
	}
	for (bit=0; bit < PB_BITS; bit++){	/* initialise timers. */
		re_at_ns[bit] = fe_at_ns[bit] = ee_at_ns[bit] = 0;
	}
	mains_phase = 2 * M_PI * rand_u();	/* phase is random wrt mains */
 	deprintf("Chip initialised with random/invalid register contents. inbus_prev set to 0x%04X.\n", inbus_prev);
}


/* Reset the lines, pointed to by the respective row[q] counter. Reset all quads together (because ResetB's are paralleled), */
void reset_lines(){
 	static int warn_count1 = 0, warn_count2 = 0, last_reset_instr = -100;
	double decay_factor;
	int q,c, new = 0;

	decay_factor = exp (- (double)length_ns/reset_tconst_ns);	/* calculate e^(-t/T) outside loop. */
	for (q=0; q < QUADS; q++){
		if (row[q] < RC_BEFORE_START){
			printf ("NOTICE: attempted to reset row %d of quad %d. This is an uninitialised state and has no effect.\n", row[q], q);
		}else if (row[q] == RC_BEFORE_START){	/* Immediately after Fsync, the row register doesn't point to anything. */
			eprintf ("NOTICE: attempted to reset row %d of quad %d. This is before-the-start and has no effect.\n", row[q], q);
		}else if (row[q] == NUM_ROWS){	/* According to the manual, it's legal to clock off-the-end, but it's stupid to try to reset that row.*/
			eprintf ("NOTICE: attempted to reset row %d of quad %d. This is 1 off-the-end and has no effect.\n", row[q], q);
		}else if (row[q] > NUM_ROWS){	/* BUT, only throw error one later, because resetting is a state (active low) rather than an edge. */
			fatal_error(ERROR_RED": attempted to reset row %d of quad %d. This is off-the-end and has no effect.\n", row[q], q);
		}else{
			for (c=0; c <NUM_COLS; c++){
				pixels_reset_ns[q][row[q]][c] = elapsed_ns + length_ns;	/* reset timestamp is the _end_ of this instruction */
				pixels_lastupdated_ns[q][row[q]][c] = elapsed_ns + length_ns; /* last updated. (also, use the _end_ of this instruction as the timestamp) */
				pixels_read_count[q][row[q]][c] = 0;			/* reset the pixel read counter */

				if (resetmode == RESET_MODE_ZERO){			/* pixel data: zero*/
					/* new = 0 already */
				}else if (resetmode == RESET_MODE_CONST){		/* pixel data: constant reset_level */
					new = reset_level;
				}else if (resetmode == RESET_MODE_NOISE){		/* pixel data: reset to constant + noise */
					new = reset_level + dither (thermal_noise (hawaii_noise));
				}else if (resetmode == RESET_MODE_EXPON){		/* pixel data: (exponentially reset towards constant) + noise */
					new = pixels[q] [row[q]] [c] - reset_level;	/* y_1 =  (y_0 - offset) * exp (- t/T)  + offset */
					new = reset_level + dither( (decay_factor * new)  + thermal_noise (hawaii_noise) );
				}

				if (new < 0){			/* Note: it is (physically) OK for the pixel voltage to go outside [0,0xffffff], even though the ADC may clip later. */
					if (warn_count1 < 10){	/* Just notice, not warning; don't clip the value. */
						eprintf ("NOTICE: (reset_level + noise) is negative. (Value is %d. This is ok, but should be rare).\n", new);
					}else if (warn_count1 == 10){
						eprintf ("NOTICE: [repeated 10 times; will not be repeated again.]\n");
					}
					warn_count1++;
					if ((pixels[q] [row[q]][c] > 0) && (new - pixels[q][row[q]][c] > 0)){	/* Detect wrap. Very unlikely; very bad. Warnings unsquelchable */
						eprintf ("WARNING: pixel reset, overflowed and wrapped around from + to -. Now %d.\n", new);
						new = 0xdeadbeef;   /* would it be more physical to use 0xffffff ? */
					}
				}else if (new > MAX_24BIT){	/* shouldn't happen */
					if (warn_count2 < 10){
						eprintf ("NOTICE: (reset_level + noise) is too high. (Value is 0x%x. This shouldn't happen, unless the noise is insanely huge.)\n", new);
					}else if (warn_count2 == 10){
						eprintf ("WARNING: [repeated 10 times; will not be repeated again.]\n");
					}
					warn_count2++;
					if ((pixels[q][row[q]][c] < 0) && (new - pixels[q][row[q]][c] < 0)){	/* Detect wrap. Very unlikely; very bad. Warnings unsquelchable */
						eprintf ("WARNING: pixel reset, underflowed and wrapped around from - to +. Now %d.\n", new);
						 new = 0xbeefdead;   /* would it be more physical to use 0 ? */
					}
				}

				pixels[q] [row[q]] [c] = new;
			}
		}
	}
	if (resetmode == RESET_MODE_ZERO){
		deprintf ("  Reset pixels: zero.\n");
	}else if (resetmode == RESET_MODE_CONST){
		deprintf ("  Reset pixels: constant %d.\n", reset_level);
	}else if (resetmode == RESET_MODE_NOISE){
		deprintf ("  Reset pixels: constant %d + noise %.2f.\n", reset_level, hawaii_noise);
	}else if (resetmode == RESET_MODE_EXPON){
		deprintf ("  Reset pixels: exponential decay to (constant %d + noise %.2f); time_const = %d ns, t = %lld ns, t/T = %.2f .\n", reset_level, hawaii_noise, reset_tconst_ns, length_ns, (float)length_ns/reset_tconst_ns);
	}
	if (!quiet){
		if (instr_count - last_reset_instr > 1){   //distinguish between a single act of reset, and a continuous/repeated action of resettting (ResetB being held low for multiple instructions).
			reset_count++;
			fprintf(stdout, "    "RESET_GREEN":  %3d,       %3d,       %3d,       %3d,         (%.3f ms)                                    \t#Q{0,1,2,3}row,    (rst_len)\n" , row[0], row[1], row[2], row[3], ((float)length_ns/NS_MS) );
		}else{
			fprintf(stdout, "  RstCont:  %3d,       %3d,       %3d,       %3d,         (%.3f ms)                                    \t#Q{0,1,2,3}row,    (rst_len)\n" , row[0], row[1], row[2], row[3], ((float)length_ns/NS_MS) );
		}
	}
	last_reset_instr = instr_count;
}


/* Illuminate the 4 pixels of interest, with various types of light source and shot noise. Note: no thermal noise here because (a) it's a capacitor, and (b) dividing the integration period into smaller dt would  change the value! */
/* Do this with "lazy_evaluation" - save the last-updated timestamp, and only calculate the value of any one pixel when it's time to read its value (or when READ_ENABLE changes). */
void lazy_illuminate_pixels(){   /* NB dither/noise are outside the brackets wrt length-multiplier. */
 	static int warn_count1 = 0, warn_count2 = 0;
	unsigned long long exposure_ns;
	int q, new=0, px;
	for (q=0; q < QUADS; q++){	/* All quads are read and reset simultaneously, though the relevant row and column may differ. */
		if ( (row[q] >= NUM_ROWS) || (row[q] < RC_BEFORE_START) || (row[q] == RC_BEFORE_START) ||  (col[q] >= NUM_ROWS) || (col[q] == RC_BEFORE_START) ){ /* Invalid positon: no physical pixel, do nothing. */
			eprintf ("NOTICE: illuminate_pixels was called for a row/col that is uninitialised, before-the-start, or off-the-end of the array. Ignored. (row=%d,col=%d).\n", row[q], col[q]);
			exposure_ns = 0;  /* keep the debug block below happy */
			continue;
		}
		exposure_ns = elapsed_ns - pixels_lastupdated_ns [q] [row[q]] [col[q]];
		if (RESETB & inbus_lo){					/* Pixels in process of reset: don't change anything. (Should have been called anyway in this case) */
			new = pixels [q] [row[q]] [col[q]];
		}else if (exposemode == EXPOSURE_MODE_DEBUG){		/* Exposure mode: debugging, own address (quad, row, col,random) */
			new = (q << 22)  +  (row[q] <<13)  +  (col[q] <<4)  +  rand_n(0x0f);
		}else if (exposemode == EXPOSURE_MODE_LINEAR){		/* Exposure mode: linear exposure to photons */
			new = pixels[q] [row[q]] [col[q]] + dither( (double)exposure_ns * light_ens );
		}else if (exposemode == EXPOSURE_MODE_GLOW){		/* Exposure mode: linear with glow  [only if read_enabled] and background-glow (always). */
			new = pixels[q] [row[q]] [col[q]]  + dither(  shot_noise( (double)exposure_ns * light_ens )  +  shot_noise( (double)exposure_ns * (glow_ens * read_enabled + background_ens) ) );
		}else if ( (exposemode == EXPOSURE_MODE_LINEARTIFF) || (exposemode == EXPOSURE_MODE_EQUALTIFF) || (exposemode == EXPOSURE_MODE_IMAGETIFF) ){ /* Equaltiff, Lineartiff, or ImageTiff */
			if (q == 0){	/* Get the correct pixel - NB different origins. Shift right from 32-bit to 24-bit */
				px = pixels_tiff [NUM_ROWS*2 -1 - row[q]] [col[q] + NUM_COLS] >>8;
			}else if (q == 1){
				px = pixels_tiff [NUM_ROWS*2 -1 - row[q]] [col[q]] >>8;
			}else if (q == 2){
				px = pixels_tiff[NUM_ROWS -1 - row[q]] [col[q]] >>8;
			}else{
				px = pixels_tiff[NUM_ROWS -1 - row[q]] [col[q] + NUM_COLS] >>8;
			}
			if (exposemode == EXPOSURE_MODE_EQUALTIFF){		/* Exposure mode: set the value exactly equal to the tiff, without integration or noise. */
				new = px;
			}else if (exposemode == EXPOSURE_MODE_LINEARTIFF){	/* Exposure mode: linear from image (scaled), without glow */
				new =  pixels [q] [row[q]] [col[q]]  +  dither(  ((double)exposure_ns/MAX_24BIT) * px * tiff_ens);
			}else{							/* Exposure mode: linear from image (scaled), with glow */
				new =  pixels[q] [row[q]] [col[q]]  +  dither(  ((double)exposure_ns/MAX_24BIT) * px  * tiff_ens +  shot_noise( (double)exposure_ns * (glow_ens * read_enabled + background_ens) ) );
			}
		}

		if (new < 0){			/* Note: it is (physically) OK for the pixel voltage to go outside [0,0xffffff], even though the ADC may clip later. */
			if (warn_count1 < 10){	/* Just notice, not warning; don't clip the value. */
				eprintf ("NOTICE: exposed pixels + noise are negative. (Value is %d. This is ok, but should be rare).\n", new);
			}else if (warn_count1 == 10){
				eprintf ("NOTICE: [repeated 10 times; will not be repeated again.]\n");
			}
			warn_count1++;
			if ((pixels[q] [row[q]] [col[q]] > 0) && (new - pixels[q] [row[q]] [col[q]] > 0)){	/* Detect wrap. Very unlikely; very bad. Warnings unsquelchable */
				eprintf ("WARNING: pixel exposed, overflowed and wrapped around from + to -. Now %d.\n", new);
				new = 0xdeadbeef;   /* would it be more physical to use 0xffffff ? */
			}
		}else if (new > MAX_24BIT){
			if (warn_count2 < 10){
				eprintf ("NOTICE: over-exposed pixels. Value is 0x%x. Will subsequently be clipped to 0x%x by ADC.\n", new, MAX_24BIT);
			}else if (warn_count2 == 10){
				eprintf ("NOTICE: [repeated 10 times; will not be repeated again.]\n");
			}
			warn_count2++;
			if ((pixels[q] [row[q]] [col[q]] < 0) && (new - pixels[q] [row[q]] [col[q]] < 0)){	/* Detect wrap. Very unlikely; very bad. Warnings unsquelchable */
				eprintf ("WARNING: pixel exposed, underflowed and wrapped around from - to +. Now %d.\n", new);
				new = 0xbeefdead;   /* would it be more physical to use 0 ? */
			}
		}
		pixels[q][row[q]][col[q]] = new;			/* new value */
		pixels_lastupdated_ns[q][row[q]][col[q]] = elapsed_ns;  /* update the timestamp, for use next time */
	}

	/* TODO: when this is called from lazy_illuminate_array(), consider squashing it to only print one line rather than 512*512 */
	if (RESETB & inbus_lo){
		deprintf ("  Lazy illuminate pixels: inactive (being reset simultaneously). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}else if (exposemode == EXPOSURE_MODE_DEBUG){
		deprintf ("  Lazy illuminate pixels: debug (value is own address, q[2]:r[9]:c[9]:random[4]) bits). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}else if (exposemode == EXPOSURE_MODE_LINEAR){
		deprintf ("  Lazy illuminate pixels: linear (%.2f electrons/ns, for %lld ns). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", light_ens, exposure_ns, row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}else if (exposemode == EXPOSURE_MODE_GLOW){
		deprintf ("  Lazy illuminate pixels: linear with glow (illumination %.2f, glow %.2f, bkground %.2f e/ns, for %lld ns). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", light_ens, glow_ens, background_ens, exposure_ns, row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}else if (exposemode == EXPOSURE_MODE_EQUALTIFF){
		deprintf ("  Lazy illuminate pixels: debug:image. (set to exact value of tiff). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}else if (exposemode == EXPOSURE_MODE_LINEARTIFF){
		deprintf ("  Lazy illuminate pixels: linear from tiff, without glow (white pixels are %.2f e/ns for %lld ns). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", tiff_ens, exposure_ns, row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}else if (exposemode == EXPOSURE_MODE_IMAGETIFF){
		deprintf ("  Lazy illuminate pixels: linear from tiff with glow (white pixels are %.2f, glow %.2f, bkground %.2f e/ns, for %lld ns). Pixels: (%d,%d, %d,%d, %d,%d, %d,%d).\n", tiff_ens, glow_ens, background_ens, exposure_ns, row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3]);
	}
}


/* Update the pixel values for the entire array. Do the "historical" exposure - this 'lazy evaluation' makes the algorithm far far faster */
void lazy_illuminate_array(){		/* This just calls lazy_illuminate_pixels for each pixel in the array */
	unsigned int row_save[QUADS], col_save[QUADS], q, r, c;
	for (q=0; q< QUADS; q++){	/* save */
		row_save[q] = row[q];
		col_save[q] = col[q];
	}
	for (r=0; r<NUM_ROWS; r++){
		for (c=0; c<NUM_COLS; c++){
			for (q=0; q< QUADS; q++){
				row[q] = r;	/* Because lazy_illuminate_pixels() actually looks at the row[q] and col[q] counters, fake them */
				col[q] = c;
			}
			lazy_illuminate_pixels();
		}
	}
	for (q=0; q< QUADS; q++){	/* restore */
		row[q] = row_save[q];
		col[q] = col_save[q];
	}
	deprintf ("  Lazy illuminate entire array.\n");
}


/* Enable readouts */
void read_enable(){
	if (read_enabled == 0){
		deprintf ("  Read enabled (on all quads). Excess glow!\n");
		lazy_illuminate_array();	/* changed state, must do the historical exposure with the old value of read_enabled, before we change our exposure conditions. */
	}
	read_enabled = 1;
}

/* Disable readouts */
void read_disable(){
	if (read_enabled == 1){
		deprintf ("  Read disabled (on all quads). Saves glow.\n");
		lazy_illuminate_array();	/* changed state, must do the historical exposure with the old value of read_enabled, before we change our exposure conditions. */
	}
	read_enabled = 0;
}

/* Fake the NI4462 trigger - just print a message. The NI doesn't actually sample on demand, so it's better to use the fake_adc to simulate. */
/* FIXME: TODO. Ideally, we could simulate the action of the NI4462, acquiring N samples started by this trigger. But we'd need to re-architect hawaiisim, and it might not be so useful anyway. */
void trigger_ni4462(){
	if (!quiet){
		if (array_key_exists(row[0], col[0], row[1],col[1], row[2],col[2], row[3],col[3]) ){  /* don't segfault !*/
			fprintf(stdout, "   "NITRIG_YELLOW": %4d,%4d  %4d,%4d  %4d,%4d  %4d,%4d    (%d,%d,%d,%d;   %.3f, %.3f, %.3f, %.3f ms)    \t#Q{0,1,2,3}row,col (read_count; exposure)\n" , row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3], pixels_read_count[0][row[0]][col[0]], pixels_read_count[1][row[1]][col[1]], pixels_read_count[2][row[2]][col[2]], pixels_read_count[3][row[3]][col[3]], ((float)(elapsed_ns - pixels_reset_ns[0][row[0]][col[0]])/NS_MS), ((float)(elapsed_ns - pixels_reset_ns[1][row[1]][col[1]])/NS_MS), ((float)(elapsed_ns - pixels_reset_ns[2][row[2]][col[2]])/NS_MS), ((float)(elapsed_ns - pixels_reset_ns[3][row[3]][col[3]])/NS_MS)    );  /* Nth is the nth read since reset on these pixels, exp_len is how long since reset. */
		}else{
			fprintf(stdout, "   "NITRIG_YELLOW": %4d,%4d  %4d,%4d  %4d,%4d  %4d,%4d    (%d,%d,%d,%d;   %.3f, %.3f, %.3f, %.3f ms)    \t#Q{0,1,2,3}row,col (read_count; exposure)\n" , row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3], -1, -1, -1, -1, -1.1, -1.1, -1.1, -1.1 );  /* bugs data */
		}
	}
}	

/* "Read" out the 4 pixels currently pointed to by the row and column registers. Write them to the output file. */
void read_quads(){	/* cam2tiff expects the little-endian bytestream thus:  Q0_L:M:H:0, Q1_L:M:H:0; Q2_L:M:H:0; Q3_L:M:H:0.  */
	static int warn_count1 = 0, warn_count2 = 0;
	int q, adc=0, ret;
	lazy_illuminate_pixels();	/* Lazy evaluation: actually do the 'historical' exposure to light now, when we actually care about the result. */
	for (q=0; q < QUADS; q++){
		if ( (row[q] >= NUM_ROWS) || (col[q] >= NUM_COLS) ){	/* It's legal to clock off the end, but it's stupid to try to read these pixels.*/
			fatal_error(ERROR_RED": attempted to read row %d, col %d, of quad %d. This is off-the-end and would read invalid data.\n", row[q], col[q], q);
			adc = 0xbadbad; /* in case of -c, will still read. */
		}else if ( (row[q] == RC_BEFORE_START) || (col[q] == RC_BEFORE_START) ){	/* Immediately after FSync/LSync, the row/col registers don't connect anything to the output. Stupid to try to read.*/
			fatal_error(ERROR_RED": attempted to read row %d, col %d, of quad %d. This is before-the-start, and would read invalid data.\n", row[q], col[q], q);
			adc = 0xbadbad00; /* in case of -c, will still read. */
		}else{
			pixels_read_count [q] [row[q]] [col[q]] ++;
			if (read_enabled){
				if (adcmode == ADC_MODE_IDEAL){		/* adc mode: ideal reads (no noise) */
					adc = pixels[q] [row[q]] [col[q]];
				}else if (adcmode == ADC_MODE_NOISE){	/* adc mode: add noise */
					adc = pixels[q] [row[q]] [col[q]] + dither( thermal_noise(adc_noise) );
				}else if (adcmode == ADC_MODE_HUM){	/* adc mode: add noise and mains hum */
					adc = pixels[q] [row[q]] [col[q]] + dither( thermal_noise(adc_noise) + mains_hum_level * cos (2 * M_PI * mains_fhz  * elapsed_ns/NS_S + mains_phase) );
				}
			}else{
				adc = 0xbad; 				/* bad - trying to read pixels with the read line disabled! */
			}
		}

		if (adc < 0){
			if (warn_count1 < 10){
				eprintf ("WARNING: (read_level + adc_noise/hum) is negative. ADC Under-range is impossible, clipping it from %d to zero.\n", adc);
			}else if (warn_count1 == 10){
				eprintf ("WARNING: [repeated 10 times; will not be repeated again.]\n");
			}else{
				sqw_count++;
			}
			warn_count1++;
			clip_count++;
			adc = 0;
		}else if (adc > MAX_24BIT){
			if (warn_count2 < 10){
				eprintf ("WARNING: (read_level + adc_noise/hum) is too high. ADC Over-range is impossible, clipping it from 0x%x to 0x%x.\n", adc, MAX_24BIT);
			}else if (warn_count2 == 10){
				eprintf ("WARNING: [repeated 10 times; will not be repeated again.]\n");
			}else{
				sqw_count++;
			}
			warn_count2++;
			clip_count++;
			adc = MAX_24BIT;
		}

		write_buf[q] = adc;
	}
	if (! read_enabled){			/* technically, this might 'work', bot only because of RC constants in the high Z-in */
		fatal_error(ERROR_RED": trying to read pixels when the READ line is not high. Would read from undriven line.\n");
	}

	/* Write data to file */
	ret = fwrite(write_buf, sizeof(write_buf[0]), (sizeof(write_buf)/sizeof(write_buf[0])), output_fh);
	if (ret != (sizeof(write_buf)/sizeof(write_buf[0]))){  /* Did we succeed in writing the entire buffer? Else trigger error and deletion of outfile below. */
		fatal_error(ERROR_RED": could not write the whole of write_buf (%d integers, %d bytes) to output file %s\n", (int)sizeof(write_buf) / (int)sizeof(write_buf[0]), (int)sizeof(write_buf), outfile);
	}

	/* Write READ stats to stdout */
	if (!quiet){
		if (array_key_exists(row[0], col[0], row[1],col[1], row[2],col[2], row[3],col[3]) ){  /* don't segfault !*/
			fprintf(stdout, "    "READ_BLUE":   %3d,%3d    %3d,%3d    %3d,%3d    %3d,%3d      (%d,%d,%d,%d;   %.3f, %.3f, %.3f, %.3f ms)    \t#Q{0,1,2,3}row,col (read_count; exposure)\n" , row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3], pixels_read_count[0][row[0]][col[0]], pixels_read_count[1][row[1]][col[1]], pixels_read_count[2][row[2]][col[2]], pixels_read_count[3][row[3]][col[3]], ((float)(elapsed_ns - pixels_reset_ns[0][row[0]][col[0]])/NS_MS), ((float)(elapsed_ns - pixels_reset_ns[1][row[1]][col[1]])/NS_MS), ((float)(elapsed_ns - pixels_reset_ns[2][row[2]][col[2]])/NS_MS), ((float)(elapsed_ns - pixels_reset_ns[3][row[3]][col[3]])/NS_MS)    );  /* Nth is the nth read since reset on these pixels, exp_len is how long since reset. */
		}else{
			fprintf(stdout, "    "READ_BLUE":   %3d,%3d    %3d,%3d    %3d,%3d    %3d,%3d      (%d,%d,%d,%d;   %.3f, %.3f, %.3f, %.3f ms)    \t#Q{0,1,2,3}row,col (read_count; exposure)\n" , row[0],col[0], row[1],col[1], row[2],col[2], row[3],col[3], -1, -1, -1, -1,  -1.1, -1.1, -1.1, -1.1 );    //bogus data.
		}
	}
	/* Debugging info: repeat ADC read values */
	if (adcmode == ADC_MODE_IDEAL){
		deprintf ("  ADC Pixels (co-ords row/y, col/x): Q0: 0x%06X (%d,%d); Q1: 0x%06X (%d,%d); Q2: 0x%06X (%d,%d);  Q3: 0x%06X (%d,%d). (ADC mode: ideal)\n", write_buf[0], row[0], col[0], write_buf[1], row[1], col[1], write_buf[2], row[2], col[2], write_buf[3], row[3], col[3]);
	}else if (adcmode == ADC_MODE_NOISE){
		deprintf ("  ADC Pixels (co-ords row/y, col/x): Q0: 0x%06X (%d,%d); Q1: 0x%06X (%d,%d); Q2: 0x%06X (%d,%d);  Q3: 0x%06X (%d,%d). (ADC mode: noise: %.2f)\n", write_buf[0], row[0], col[0], write_buf[1], row[1], col[1], write_buf[2], row[2], col[2], write_buf[3], row[3], col[3], adc_noise);
	}else if (adcmode == ADC_MODE_HUM){
		deprintf ("  ADC Pixels (co-ords row/y, col/x): Q0: 0x%06X (%d,%d); Q1: 0x%06X (%d,%d); Q2: 0x%06X (%d,%d);  Q3: 0x%06X (%d,%d). (ADC mode: noise: %.2f, hum %.2f)\n", write_buf[0], row[0], col[0], write_buf[1], row[1], col[1], write_buf[2], row[2], col[2], write_buf[3], row[3], col[3], adc_noise, mains_hum_level);
	}
	read_count++;
}

/* Clock PIXEL[quad q].  Check for overflow. [This should happen on either edge of PIXEL_n.] */
void clock_pixel(int q){	/* Note: the physical chip disconnects the column register after 511, then "falls off the end". We consider 512 as legal, 514+ as obvious errors. (Manual doesn't specify, but this sets Linechk). */
	col[q]++;		/* Rockwell confirm that the register "goes off the end" rather than wrapping */
	deprintf ("  Pixel clock (quad %d) incremented, now %d.\n", q, col[q]);
	if ( (col[q] == NUM_COLS) || (col[q] == NUM_COLS+1) ){  //off the end by one or two. Legal, be quiet unless debug.
		deprintf ("  Pixel clock (quad %d) now 1 or 2 off the end, at column %d.\n", q, col[q]);
	}else if (col[q] >= NUM_COLS +2){  /* Hardware would just "go off the end" harmlessly, without wrapping. But we should detect an error if this happens */
		fatal_error(ERROR_RED": Pixel clock (quad %d) is now at %d, which exceeds number of columns (%d) by 2. Legal, but strongly suggests a bug. (pc: %d, source_line: %d).\n", q, col[q], NUM_COLS, instr_count, line_num);
	}
}

/* Clock LINE[quad q].  Check for overflow. [This should happen on either edge of LINE_n.] */
void clock_line(int q){		/* Note: it may be clocked to > 511 , but shouldn't be used when "off the end". (Manual doesn't specify, but this sets Frmchk).  */
	row[q]++;
	deprintf ("  Line clock (quad %d) incremented, now %d.\n", q, row[q]);
	if ( (row[q] == NUM_ROWS) || (row[q] == NUM_ROWS+1) ){  //off the end by one or two. Legal, be quiet unless debug
		deprintf ("  Line clock (quad %d) now 1 or 2 off the end, at column %d.\n", q, row[q]);
	}else if (row[q] >= NUM_ROWS +2){			//off by more than 2; strongly suggestive of bug.
		fatal_error(ERROR_RED": Line clock (quad %d) is now at %d, which exceeds number of rows (%d) by 2. Legal, but strongly suggests a bug. (pc: %d, source_line: %d).\n", q, row[q], NUM_ROWS, instr_count, line_num);
	}
}

/* Do LSYNC[*]. Reset the pixel (column) counter. Check that we meet the condition for Pixel[*] being low when Lsync is low. [This should happen whenever LSYNC is low.] */
void lsync(){
	int q;
	for (q=0; q < QUADS; q++){
		col[q] = RC_BEFORE_START;   /* Note that after LSync, NO column is connected to the output. Represent this as -1 */
	}
	deprintf ("  Lsync done (all quads). Cols set to %d.\n", RC_BEFORE_START);
}

/* Do FSYNC[*]. Reset the line (row) counter. Check that we meet the condition for Line[*] being low when Fsync is low. [This should happen whenever FSYNC is low.] */
void frsync(){		/* call this function frsync(), since fsync() is already defined by glibc. */
	int q;
	for (q=0; q < QUADS; q++){
		row[q] = RC_BEFORE_START;   /* Note that after FSync, NO row is connected to the output. Represent this as -1 */
	}
	deprintf ("  Fsync done (all quads). Rows set to %d.\n", RC_BEFORE_START);
}


/* Do one step of simulation, based on the input data word (inbus) and the runtime (length_ns); */
void simulate_step(){
	int bit, wbit, level, ferr_count = 0;
	char *txt1, *txt2;	/* static text */
	unsigned long long pulse_duration[PB_BITS]; /* duration of current pulse on that bit. */

	if (inbus & ~USED){		/* Check input lines are all valid */
 		fatal_error(ERROR_RED": input bus was 0x%06X, but only these bits are used: 0x%06X. Unused lines are asserted in error. (pc: %d, source_line: %d).\n",inbus, USED, instr_count, line_num);
 	}

  	inbus_hi = inbus;			/* High bits */
  	inbus_lo = USED & ~inbus;		/* Low bits (bits that are low in input are set in inbus_lo)*/
  	inbus_re  = inbus_hi & ~inbus_prev;	/* Bits that just went high (rising edge) */
  	inbus_fe  = inbus_lo & inbus_prev;	/* Bits that just went high (rising edge) */
  	inbus_ee  = inbus ^ inbus_prev;		/* Bits that toggled (either edge) */
 	// eprintf ("Step: %d, inbus_prev: 0x%x, inbus: 0x%x, inbus_hi: 0x%x, inbus_lo: 0x%x, inbus_re: 0x%x, inbus_fe: 0x%x, inbus_ee: 0x%x.\n", instr_count, inbus_prev, inbus, inbus_hi, inbus_lo, inbus_re, inbus_fe, inbus_ee);

 	/* Deal with each line of the bus's action. */
  	if (PIXEL_0 & inbus_ee) { clock_pixel(0); }	/* Clock pixel 0 on either edge of PIXEL_0 */
  	if (PIXEL_1 & inbus_ee) { clock_pixel(1); }	/* Clock pixel 1 on either edge of PIXEL_1 */
  	if (PIXEL_2 & inbus_ee) { clock_pixel(2); }	/* Clock pixel 2 on either edge of PIXEL_2 */
  	if (PIXEL_3 & inbus_ee) { clock_pixel(3); }	/* Clock pixel 3 on either edge of PIXEL_3 */

  	if (LINE_0  & inbus_ee) { clock_line(0);  }	/* Clock line 0 on either edge of LINE_0 */
  	if (LINE_1  & inbus_ee) { clock_line(1);  }	/* Clock line 1 on either edge of LINE_1 */
  	if (LINE_2  & inbus_ee) { clock_line(2);  }	/* Clock line 2 on either edge of LINE_2 */
  	if (LINE_3  & inbus_ee) { clock_line(3);  }	/* Clock line 3 on either edge of LINE_3 */

  	if (LSYNC   & inbus_lo) { lsync();	  }	/* Do Lsync if LSYNC is low */
  	if (FSYNC   & inbus_lo) { frsync();	  }	/* Do Fsync if FSYNC is low */

  	if (READ    & inbus   ) { read_enable();  }     /* Enable(disable) the Read lines (all pixels) if READ is high(low) */
 	if (READ    & inbus_lo) { read_disable(); }

 	if (ADC & inbus_fe)     { trigger_ni4462();}	/* Do a virtual ADC trigger on falling edge. This is the NI4462 begin_acquisition, we can't ACTUALLY emulate the real reads here. */

	if (FAKEADC & inbus_fe) { read_quads();   }	/* Do fake-ADC trigger on falling edge. Emulate a single ACTUAL read */

 	if (RESETB  & inbus_lo) { reset_lines();  }     /* Do ResetB (reset all pixels in active row) if RESETB is low.  NB do this AFTER the ADC, because this is active low, rather than edge. */


	/* First instruction: fake the edges. Otherwise, we will often get a fatal-error when we try to initialise. */
	if (instr_count == 0){
		deprintf ("First instruction fudge: removed edges before check. (Otherwise, can't initialise).\n");
		inbus_ee = inbus_re = inbus_fe = 0;
	}

	/* Check the requirements between the various clocks */
	if ( (LSYNC & inbus_re) && ( (PIXELS & inbus) || (PIXELS & inbus_fe) ) ){  /* Important: Lsync is active low; as it goes high, mustn't trigger pixel-clocks */
		eprintf(ERROR_RED": Lsync rising edge coincides with a high (or falling edge) Pixel-clock. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (LSYNC & inbus_lo) && ( (PIXELS & inbus) || (PIXELS & inbus_fe) ) ){  /* Documented: "Pixel clock must be low when Lsync is pulsed low" (Yes, this is a strict requirement; they can't fall together) */
		eprintf(ERROR_RED": Lsync low coincides with a high (or falling edge) Pixel-clock. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (LSYNC & inbus_fe) && ( (PIXELS & inbus) || (PIXELS & inbus_fe) ) ){  /* Documented: "Pixel clock must be low when Lsync is pulsed low" (Yes, this is a strict requirement; they can't fall together) */
		eprintf(ERROR_RED": Lsync falling edge coincides with a high (or falling edge) Pixel-clock. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}

	if ( (FSYNC & inbus_re) && ( (LINES & inbus) || (LINES & inbus_fe) ) ){  /* Important: Fsync is active low; as it goes high, mustn't trigger line-clocks */
		eprintf(ERROR_RED": Fsync rising edge coincides with a high (or falling edge) Line-clock. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (FSYNC & inbus_lo) && ( (LINES & inbus) || (LINES & inbus_fe) ) ){  /* Documented: "Line clock must be low when Fsync is pulsed low" (Yes, this is a strict requirement; they can't fall together)  */
		eprintf(ERROR_RED": Fsync low coincides with a high (or falling edge) Line-clock. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (FSYNC & inbus_fe) && ( (LINES & inbus) || (LINES & inbus_fe) ) ){  /* Documented: "Line clock must be low when Fsync is pulsed low" (Yes, this is a strict requirement; they can't fall together) */
		eprintf(ERROR_RED": Fsync falling edge coincides with a high (or falling edge) Line-clock. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}

	if ( (RESETB & inbus_ee) && ( (PIXELS|LINES) & inbus_ee) ) {		  /* Avoid glitches when ResetB and another Line/Pixel clock toggle together. */
		eprintf(ERROR_RED": ResetB edge coincides with an edge on Pixel-clock or Line-clock. Glitch potential. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}

	/* These are useful for "FAKEADC; the real ADC has a trigger-delay, which makes all these tests irrelevant */
	if ( (FAKEADC & inbus_fe) &&  ( (PIXELS|LINES) & inbus_ee) ) {		 /* Avoid glitches: ADC trigger (FE) must not coincide with  Line or Pixel clock. */
		eprintf(ERROR_RED": ADC falling edge coincides with an edge on Pixel-clock or Line-clock. Glitch potential. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (FAKEADC & inbus_fe) &&  ( (LSYNC|FSYNC) & inbus_lo) ){		/* Avoid triggering ADC while Lsync/Fsync low. This just seems like a bad idea. */
		eprintf(ERROR_RED": ADC falling edge coincides with LSync or FSync low. Bad idea. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (FAKEADC & inbus_fe) &&  (RESETB & inbus_ee)  ){			/* Avoid ADC trigger simultanously with edge on ResetB. Glitch potential. */
		eprintf(ERROR_RED": ADC falling edge coincides with ResetB edge. Glitch potential. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (FAKEADC & inbus_fe) &&  (RESETB & inbus_lo)  ){			/* Avoid ADC trigger simultanously with ResetB low. Seems like a bad idea. */
		eprintf(ERROR_RED": ADC falling edge coincides with ResetB low. Bad idea. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}
	if ( (FAKEADC & inbus_fe) &&  ( (READ & inbus_lo) || (READ & inbus_re )  ) ){ /* ADC trigger would be bogus if READ is low, and glitchy if read just had a rising edge. */
		eprintf(ERROR_RED": ADC falling edge coincides with low (or rising edge) Read. (inbus: 0x%04X, inbus_prev: 0x%04X, pc: %d, source_line: %d).\n",inbus, inbus_prev, instr_count, line_num);
		ferr_count++;
	}


	/* Calculate pulse lengths: for each bit, work out how long it has been in the present state. */
	for (bit=0; bit < PB_BITS; bit++){	/* For each bit, how long has the current/previous pulse lasted since the last edge. (if we are on edge now, count prev duration, NOT zero) */
		pulse_duration[bit] = elapsed_ns - ee_at_ns[bit];
	}

	/* Check pulse lengths are sufficiently long (also setup/hold times) */
	for (bit=0; bit < PB_BITS; bit++){	/* For each bit... */
		wbit = 0x1 << bit;		/* [word with that bit set] */
		if (wbit & inbus_ee){		/*  ...only look at bits that changed */
			if (wbit & inbus_re){
				txt1 = "low";		/* before */
				txt2 = "rising";	/* edge */
				level = 1;		/* level now */
			}else{
				txt1 = "high";
				txt2 = "falling";
				level = 0;
			}
			deprintf ("  Bit %d changed (%s edge), duration of previous %s was %lld ns (%.2f ms)\n", bit, txt2, txt1, pulse_duration[bit], ((float)pulse_duration[bit]/NS_MS) );
			if (instr_count > 0){		/* squelch error on first instruction */

				if ( (POSN_CLOCKS & wbit) &&  (pulse_duration[bit] < MIN_PULSE_POSN_NS) ){	/* Check pulse length is sufficient for the positioning clocks */
					eprintf(ERROR_RED": Pulse duration on position clock (bit %d %s for %lld ns, just finished) was too short (min_pulse_ns is %d ns). (source_line %d).\n", bit, txt1, pulse_duration[bit], MIN_PULSE_POSN_NS, line_num);
					ferr_count++;
				}else if ( (RESETB & wbit & inbus_re) &&  (pulse_duration[bit] < MIN_PULSE_RESET_NS) ){	/* Check length is sufficient for the RESET */
					eprintf(ERROR_RED": Pulse duration on reset (bit %d %s for %lld ns, just finished) was too short (min_pulse_ns is %d ns). (source_line %d).\n", bit, txt1, pulse_duration[bit], MIN_PULSE_RESET_NS, line_num);
					ferr_count++;
				}else if  ( (LINES & wbit) && (pulse_duration[bit] < MIN_PULSE_RESET_NS) && ( RESETB & ((inbus_lo & ~inbus_fe) | inbus_re) ) ){ /* Check that each LINE is reset for long enough. Bad if line pulse is short, and ResetB is (low but not fe) or re */
					eprintf(ERROR_RED": Pulse duration on line clock (bit %d %s for %lld ns, just finished) was too short for a Reset (min_pulse_ns is %d ns). (source_line %d).\n", bit, txt1, pulse_duration[bit], MIN_PULSE_RESET_NS, line_num);
					ferr_count++;
				}else if ( (ADC & wbit) && (pulse_duration[bit] < MIN_PULSE_ADC_NS) ){		/* Check length is sufficient for the (real) ADC (min pulse length) */
					eprintf(ERROR_RED": Pulse duration on ADC (bit %d %s for %lld ns, just finished) was too short (min_pulse_ns is %d ns). (source_line %d).\n", bit, txt1, pulse_duration[bit], MIN_PULSE_ADC_NS, line_num);
					ferr_count++;
				}else if ( (FAKEADC & wbit) && (level == 1) && (pulse_duration[bit] < MIN_PULSE_FAKEADC_NS) ){	/* Check length is sufficient for the (fake) ADC (min pulse length) */
					eprintf(ERROR_RED": Pulse duration on Fake ADC (bit %d %s for %lld ns, just finished) was too short (min_pulse_ns is %d ns). (source_line %d).\n", bit, txt1, pulse_duration[bit], MIN_PULSE_FAKEADC_NS, line_num);
					ferr_count++;					
				}else if ( (FAKEADC & wbit) && (level == 0) && ((elapsed_ns - fe_at_ns[FAKEADC_BIT]) < MIN_INTERVAL_FAKEADC_NS) ){	/* ADC fe..fe time (retrigger interval, this is the important one.)] */
					eprintf(ERROR_RED": Fake ADC retrigger interval (falling-edge...falling-edge) (bit %d for %lld ns) was too short (min_interval_ns is %d ns). (source_line %d).\n", bit, (elapsed_ns - fe_at_ns[FAKEADC_BIT]), MIN_INTERVAL_FAKEADC_NS, line_num);
						ferr_count++;
				}else if ( (FAKEADC & wbit) && (level == 0) && (pulse_duration[READ_BIT] < MIN_SETUP_READ_FAKEADC_NS) ){	/* Check Read setup time (high) vs ADC trigger (fe). [Already verified that read is high, see above] */
					eprintf(ERROR_RED": Setup time (Read high) before Fake ADC trigger (bit %d has been high for %lld ns) was too short (min_pulse_ns is %d ns). (source_line %d).\n", READ_BIT, pulse_duration[READ_BIT], MIN_SETUP_READ_FAKEADC_NS, line_num);
					ferr_count++;
				}else if ( (READ & wbit) && (level == 0) && (FAKEADC & inbus_lo) && (pulse_duration[FAKEADC_BIT] < MIN_HOLD_READ_FAKEADC_NS) ){ /* Check Read hold time (high) vs ADC trigger (fe). [Look at READ fe, vs how long ADC has been low] */
					eprintf(ERROR_RED": Hold time (Read high) after Fake ADC trigger (bit %d has been low for %lld ns) was too short (min_pulse_ns is %d ns). (source_line %d).\n", FAKEADC_BIT, pulse_duration[FAKEADC_BIT], MIN_HOLD_READ_FAKEADC_NS, line_num);
					ferr_count++;
				}
			}
		}
	}

	/* Update pulse lengths */
	for (bit=0; bit < PB_BITS; bit++){	/* For each bit, set new timestamps if we just saw an edge */
		wbit = 0x1 << bit;
		if (wbit & inbus_re){
			re_at_ns[bit] = elapsed_ns;
			ee_at_ns[bit] = elapsed_ns;
		}else if (wbit & inbus_fe){
			fe_at_ns[bit] = elapsed_ns;
			ee_at_ns[bit] = elapsed_ns;
		}
	}

	/* Did at least one fatal error occur? */
	if (ferr_count > 0){
		fatal_error2();
	}

	/* Expose it to some light! */
	/* Used to call illuminate_array() here. But that's very algorithmically inefficient. Integration really is the sum of small steps, so use lazy_evaluation when needed instead */

	/* End instruction; update; next */
	elapsed_ns += length_ns;
	inbus_prev = inbus;
	instr_count++;
}


void gen_cam(){	/* Simple, test function. This should generate a perfect .cam format file. */
	/* Note: since ADC takes 1.3 secs for full-frame, #define overall exposure to ~ 10 secs i.e. >> ADC_len */
	int r,c;

	eprintf("Test mode: reset/illuminate/read the entire array; get a full-frame image as output. (Reset for %d ns, expose for %.2f ms, ADC for %.2f ms.)\n", gencam_reset_ns, (double)gencam_exposure_ns/NS_MS, (double)MIN_INTERVAL_FAKEADC_NS*NUM_ROWS*NUM_COLS/NS_MS);

	length_ns = gencam_reset_ns; 		/* A bit less than 1 us, for reset. */
	frsync();				/* Reset */

	clock_line(0); clock_line(1); clock_line(2); clock_line(3); /* clock to the zeroth row */
	for (r=0; r <NUM_ROWS; r++){
		reset_lines();
		clock_line(0); clock_line(1); clock_line(2); clock_line(3);
	}
	elapsed_ns += length_ns;

	read_enable();
	elapsed_ns += gencam_exposure_ns; 	/* A bit less than 10 sec, for average of nearly full exposure. */

	frsync();
	clock_line(0); clock_line(1); clock_line(2); clock_line(3); /* clock to the zeroth row */

	for (r=0; r<NUM_ROWS; r++){		/* Read, and outputs all pixels in the .cam raster. */
		lsync();
		clock_pixel(0); clock_pixel(1); clock_pixel(2); clock_pixel(3); /* clock to the zeroth column */
		for (c=0; c< NUM_COLS; c++){
			read_quads();			/* implies doing lazy_illumination() for this pixel first */
			elapsed_ns += MIN_INTERVAL_FAKEADC_NS;   /* although one might want a 'hyperfast adc' that can read all pixels at the same time, that would break the '-a hum' feature. */
			clock_pixel(0); clock_pixel(1); clock_pixel(2); clock_pixel(3);
		}
		clock_line(0); clock_line(1); clock_line(2); clock_line(3);
	}
}


/* Parse a line (line_num) of source, (passed in buffer), which should be in the (trivial) pbsim format. Put the result into inbus, length_ns. */
int parse_sourceline (char buffer[BUFFER_LEN], int line_num){		/* Return -1 for comment/blank-line, 0 on success; 1 on parse-error. */

	/* Example pbsim log format, as documented in /usr/local/share/doc/pb_parse/pbsim.txt:
	 *
	 *    //This log file was generated by pb_parse from source file source.pbsrc on date 2011-01-12.
	 *    //File format: output (uint32), length (uint64). Note that length is in ns, not ticks.
	 *    0x000001        1000000000
	 *    0x000002        1000000000
	 *    0x000003        1000000000
	 *    0x000004        1000000000
	 *
	 *  Note that "wait" opcodes (see pbsim.txt) get no special treatment, nor is one really needed. (though they are detected and result in a WARNING).
        */

	char buf_copy[BUFFER_LEN];	/* Copy of the source-line, because strtok chops it up in place.*/
	unsigned int column = 0;	/* which column are we on */
	char *token;   			/* pointer to string token */
	char *whitespace = "\t\n ";	/* space, newline or tab */
	char *end;			/* pointer to unused part in strtol() */

	/*If the actual line length (including the trailing \n) > BUFFER_LEN -1, then we are in trouble, since fgets will start to break up lines. */
	/*The amount read into the buffer must either be < (BUFFER_LEN -1), or, IFF the last character is a \n,  exactly BUFFER_LEN -1 */
	if ((strlen (buffer) > BUFFER_LEN - 1) && (buffer[BUFFER_LEN - 2] != '\n') ){
		eprintf ("Error in source file at line %d: Line length is too long (> %d characters). Line begins:\n\t%s\n", line_num, BUFFER_LEN - 1 ,buffer);
		return (1);
	}
	strcpy(buf_copy, buffer);	/* Safe: we defined these the same size */

 	token = strtok (buf_copy, whitespace);		/* Find the first whitespace-delimited token within the string. If the string is an empty line, nothing will be found. */
 	while (token != NULL){
		if ((token[0] == '/') && (token[1] == '/')) {  //comment
			break;
 		}
 		/* Parse the data. strtoul() converts the string (which may be decimal or hex) to unsigned long. Catch anything illegal. */
 		if (token[0] == '-'){
 			if (strlen(token) > 1){
 				eprintf("Error in source at line %d: column %d, '%s' is negative. Line is:\n\t%s\n", line_num, column +1, token, buffer);
 				return (1);
 			}
 		}else if ((token[0] == '0') && (token[1] != 'x')){ /* Why, why, why is leading zero without 0x interpreted as Octal?  Any accidental use of Octal is a human error. */
 			eprintf("Error in source at line %d: column %d, '%s' begins '0' (but not '0x'). Octal is evil! Line is:\n\t%s\n", line_num, column +1, token, buffer);
 			return (1);
		}
 		errno = 0;
 		if (column == 0){			/* Column 0, is inbus (pulseblaster output). */
 			inbus = strtoul(token,&end,0);
 		}else if (column == 1){			/* Column 1, is length_ns (time the instruction runs for) */
 			length_ns = strtoull(token,&end,0);
 		}else if (column > 1){
 			eprintf ("Error in source at line %d: too many columns. Expect 2. Line is:\n\t%s\n", line_num, buffer);
 			return (1);
 		}
 		if (errno != 0 || *end != 0 || end == token){	/* Were there any unused chars in the string, eg trailing garbage. */
 			eprintf("Error in source at line %d: strtoul()/strtoull() couldn't parse column %d, '%s'. Line is:\n\t%s\n", line_num, column +1, token, buffer);
 			return (1);
 		}
		token = strtok (NULL, whitespace); 	/* Find any subsequent whitespace-delimited token within the string */
		column++;
 	}
 	if ((column > 0) && (column < 2)){		/* Too few columns in this line! */
 		eprintf("Error in source at line %d: too few columns. Expect 2. Line is:\n\t%s\n", line_num, buffer);
 		return (1);
 	}
 	if (column == 0){				/* Result. */
		//eprintf ("Line %d parsed as: Blank line (or Comment)\n", line_num);
		return (-1);
	}else{
 		//eprintf ("Line %d parsed as: input_bus: 0x%04x,  length_ns: %02lld\n", line_num, inbus, length_ns);
		if (length_ns == 0){			/* Detect wait opcodes. But don't do anything other than warn (changeme?) */
			eprintf ("Note: WAIT opcode at line %d (it has length_ns == 0). Not treating it specially.\n", line_num);
		}
		return (0);
	}
}


/* Main program starts here... */
int main(int argc, char *argv[]){
	int  opt; extern char *optarg; extern int optind, opterr, optopt;  			/* getopt */
	char *strtod_left;									/* anything left over by strtod */
	char *resetarg=NULL, *exposearg=NULL, *adcarg=NULL;					/* mode for reset/expose/adc */
	struct timespec start_time, stop_time;	double run_time;				/* times*/
	struct stat stat_p;									/* pointer to stat structure */
	FILE *input_fh=NULL;									/* input file */
	char infile[FILENAME_LEN]="", tifffile[FILENAME_LEN]="";				/* file names */
	int  allow_overwrite = 0;								/* -x to overwrite existing file. */
	char buffer[BUFFER_LEN];								/* input file line buffer */
	TIFF *tiff=NULL;									/* image */
	int  tiff_rows, tiff_cols; uint16_t tiff_bps, tiff_spp, tiff_fillorder, tiff_orientation; /* tiff properties */
	int  r, ret, len;									/* misc */

	/* Get start time */
	clock_gettime( CLOCK_REALTIME, &start_time);
	srand (start_time.tv_nsec);	/* Seed random number generator. For "repeatable randomness", use srand(0) instead. */

	/* Defaults, which MAY be changed by options */
	reset_level 	=  RESET_LEVEL;
	reset_tconst_ns =  (double)RESET_TCONST_US * NS_US;
	glow_ens   	=  (double)GLOW_EUS / NS_US;
	background_ens  =  (double)BACKGROUND_EUS / NS_US;
	light_ens  	=  (double)LIGHT_EUS / NS_US;
	tiff_ens   	=  (double)TIFF_EUS / NS_US;
	hawaii_noise    =  HAWAII_C_NOISE;
	adc_noise	=  ADC_R_NOISE;
	gencam_reset_ns = (double)GENCAM_RESET_US * NS_US;
	gencam_exposure_ns = (long long)GENCAM_EXPOSURE_US * NS_US;
	mains_fhz	=  MAINS_FREQ_HZ;
	mains_hum_level	=  MAINS_HUM_AMPLITUDE;

	/* Get options and check for validity */
	if ((argc > 1) && (!strcmp (argv[1], "--help"))) {	/* Support --help, without the full getopt_long */
		print_help();
		exit (EXIT_SUCCESS);
	}

	while ((opt = getopt(argc, argv, "cdhqtxa:e:i:o:r:A:B:E:F:G:H:I:L:R:T:V:W:")) != -1) {	/* Getopt */
		switch (opt) {
			case 'a':				/* ADC mode */
				adcarg = optarg;
				break;
			case 'c':
				conterr = 1;			/* continue on error */
				break;
			case 'd':				/* debug on */
				debug = 1;
				break;
			case 'e':				/* Exposure (integrate) mode */
				exposearg = optarg;
				break;
			case 'h':				/* help */
				print_help();
				exit(EXIT_SUCCESS);
				break;
			case 'i':				/* infile */
				strncpy(infile, optarg, (sizeof(outfile) - 1));
				break;
			case 'o':				/* outfile */
				strncpy(outfile, optarg, (sizeof(outfile) - 1));
				break;
			case 'q':				/* make it quieter */
				quiet = 1;
				break;
			case 'r':				/* reset mode */
				resetarg = optarg;
				break;
			case 't':				/* test mode */
				test = 1;
				break;
			case 'x':				/* allow overwrite of existing file */
				allow_overwrite = 1;
				break;
			case 'A':				/* ADC_noise */
				adc_noise = strtod(optarg, &strtod_left);
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'B':				/* dummy */
				background_ens = strtod(optarg, &strtod_left) / NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'E':				/* Gencam exposure length */
	 			gencam_exposure_ns = strtod(optarg, &strtod_left) * NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'F':				/* Gencam reset length */
				gencam_reset_ns = strtod(optarg, &strtod_left) * NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'G':				/* dummy */
				glow_ens = strtod(optarg, &strtod_left) / NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'H':				/* Hawaii noise */
				hawaii_noise = strtod(optarg, &strtod_left);
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'I':				/* Image (tiff) gain */
				tiff_ens = strtod(optarg, &strtod_left) / NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'L':				/* Light amount */
				light_ens = strtod(optarg, &strtod_left) / NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'R':				/* Reset level */
				reset_level = strtoul(optarg, &strtod_left, 0);
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not integer.\n", optarg, opt);
				}
				break;
			case 'T':				/* Reset time constant */
				reset_tconst_ns = strtod(optarg, &strtod_left) * NS_US;
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'V':				/* Mains hum amplitude */
				mains_hum_level = strtod(optarg, &strtod_left);
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			case 'W':				/* Mains hum frequency (Hz) */
				mains_fhz = strtod(optarg, &strtod_left);
				if (*strtod_left != 0){
					feprintf("Error: argument %s to option -%c is not numeric.\n", optarg, opt);
				}
				break;
			default:				/* Catches wrong options or options missing their arg */
				feprintf("Error: unrecognised option -%c, use -h for help.\n", opt);
		}
	}

	if (test){			/* Test mode conflicts with -i (infile would be ignored, so explicitly ban it.) */
		if (strlen(infile) !=0){
			feprintf(ERROR_RED": option -t conflicts with -i. Input file not used in test mode.\n");
		}
	}else{
		if (strlen(infile) == 0){	/* input file must be specified, must have the .pbsim extension, must exist, must not be empty. Then open it. */
			feprintf(ERROR_RED": input file must be specified with: -i infile.%s\n", EXTN_PBSIM);
		}else if (( strcmp ( (infile + strlen(infile) - strlen(EXTN_PBSIM)), EXTN_PBSIM) ) && (!allow_overwrite) ){  /* -x skips this check */
			feprintf(ERROR_RED":, input file %s is not a .%s file. (Wrong extension, override with -x).\n", infile, EXTN_PBSIM);
		}else if (stat (infile, &stat_p) == -1){
			feprintf(ERROR_RED": input file %s does not exist\n", infile);
		}else if (stat_p.st_size == 0){
			feprintf(ERROR_RED": input file %s is empty! This usually occurs if pb_parse found a fatal error in the .pbsrc file.\n", infile);
		}else if ((input_fh = fopen(infile,"r")) == NULL){
			feprintf(ERROR_RED": could not open input file %s.\n",infile);
		}
	}

	if (strlen(outfile) == 0){	/* Outfile should normally be specified, have the .dat/.cam extension, must not exist. (If no file specified, and not -t, use infile, with changed extn). Open it for writing */
		if (test){
			feprintf(ERROR_RED": output file must be specified with -o file.%s in test (-t) mode.\n", EXTN_CAM);
		}else{
			len = strlen (infile) - strlen(EXTN_PBSIM);  /* No file specified; rename input file with output extn */
			if (abs(sizeof(outfile)) -1 < len + strlen(EXTN_DAT)){
				feprintf("Filename too long: %s\n", infile);
			}
			strncpy(outfile, infile, len );
			outfile[len] = 0;
			strncat(outfile, EXTN_DAT, (sizeof(outfile) - 1));
		}
	}else if ( (strcmp ( (outfile + strlen(outfile) - strlen(EXTN_DAT)), EXTN_DAT) ) && (strcmp ( (outfile + strlen(outfile) - strlen(EXTN_CAM)), EXTN_CAM) )  && (strcmp (outfile, DEV_NULL)) && (!allow_overwrite)  ){
		feprintf(ERROR_RED": output file %s is not a .%s (or %s) file. (Wrong extension, override with -x).\n", outfile, EXTN_DAT, EXTN_CAM); /* -x skips this check */
	}
	if ((stat (outfile, &stat_p) != -1) && (strcmp (outfile, DEV_NULL)) && (!allow_overwrite) ){  	/* Check that the dest file CAN'T be stat()ed. If it exists, then quit (except for /dev/null, or -x) */
		feprintf(ERROR_RED": output file %s already exists; will not overwrite. (override with -x or %s).\n", outfile, DEV_NULL);
	}
	if ((output_fh = fopen(outfile,"w")) == NULL){
		feprintf(ERROR_RED": could not open destination file %s for writing.\n", outfile);
	}
	if (resetarg == NULL){			/* Reset mode. Default to 'zero' */
		resetmode = RESET_MODE_ZERO;
	}else if (!strcmp (resetarg, "zero")){
		resetmode = RESET_MODE_ZERO;
	}else if (!strcmp (resetarg, "const")){
		resetmode = RESET_MODE_CONST;
	}else if (!strcmp (resetarg, "noise")){
		resetmode = RESET_MODE_NOISE;
	}else if (!strcmp (resetarg, "expon")){
		resetmode = RESET_MODE_EXPON;
	}else{
		feprintf(ERROR_RED": invalid reset_mode: %s. (Use: zero|const|noise|expon)\n", resetarg);
	}

	if (exposearg == NULL){			/* Exposure mode. Default to 'uniform' */
		exposemode = EXPOSURE_MODE_LINEAR;
	}else if (!strcmp (exposearg, "debug")){
		exposemode = EXPOSURE_MODE_DEBUG;
	}else if (!strcmp (exposearg, "uniform")){
		exposemode = EXPOSURE_MODE_LINEAR;
	}else if (!strcmp (exposearg, "glow")){
		exposemode = EXPOSURE_MODE_GLOW;
	}else if (!strncmp (exposearg, "debug:file", 6)){			/* strNcmp: just check the first 6 chars */
		exposemode = EXPOSURE_MODE_EQUALTIFF;
		strncpy(tifffile, (exposearg +6), (sizeof(tifffile) - 1));	/* after the 'debug:' */
	}else if (!strncmp (exposearg, "linear:file", 7)){			/* strNcmp: just check the first 7 chars */
		exposemode = EXPOSURE_MODE_LINEARTIFF;
		strncpy(tifffile, (exposearg +7), (sizeof(tifffile) - 1));	/* after the 'linear:' */
	}else if (!strncmp (exposearg, "image:file", 6)){ 			/* strNcmp: first 6 chars */
		exposemode = EXPOSURE_MODE_IMAGETIFF;
		strncpy(tifffile, (exposearg +6), (sizeof(tifffile) - 1));	/* after the 'tiff:' */
	}else{
		feprintf(ERROR_RED": invalid expose_mode: %s. (Use: debug|uniform|glow|debug:file|linear:file|image:file)\n", exposearg);
	}
	if ( (exposemode == EXPOSURE_MODE_EQUALTIFF) || (exposemode == EXPOSURE_MODE_LINEARTIFF) || (exposemode == EXPOSURE_MODE_IMAGETIFF) ){
		if (strlen(tifffile) == 0){
			feprintf(ERROR_RED": tiff file can not be empty string.\n");
		}else if (strcmp ( (tifffile + strlen(tifffile) - strlen(EXTN_TIFF)), EXTN_TIFF) ){
			feprintf(ERROR_RED":, image file %s is not a .%s file. (Wrong extension).\n", tifffile, EXTN_TIFF);
		}else if (stat (tifffile, &stat_p) == -1){
			feprintf(ERROR_RED": image file %s does not exist\n", tifffile);
		}else if (stat_p.st_size == 0){
			feprintf(ERROR_RED": image file %s is empty! (It should contain 32-bit pixel data for %d x %d x%d pixels.)\n", tifffile, QUADS, NUM_ROWS, NUM_COLS);
		}

		/* Open the tiff file, then check that columns.rows/bitdepth are as we expect (see cam2tiff). Then read it into pixels_tiff[][] */
		if( ( tiff = TIFFOpen ( tifffile, "r" ) ) == NULL ) {
			feprintf (ERROR_RED": could not open image file %s: %s\n", tifffile, strerror(errno) ); /* read tiff tags. c.f. cam2tiff */
		}else if ( (!TIFFGetField ( tiff, TIFFTAG_IMAGEWIDTH, &tiff_cols))   	   ||	/* Expect 1024 */
			   (!TIFFGetField ( tiff, TIFFTAG_IMAGELENGTH, &tiff_rows))  	   ||	/* Expect 1024 */
			   (!TIFFGetField ( tiff, TIFFTAG_BITSPERSAMPLE, &tiff_bps)) 	   ||	/* Expect (8 * 32) */
			   (!TIFFGetField ( tiff, TIFFTAG_SAMPLESPERPIXEL, &tiff_spp))     ||	/* Expect 1 */
			   (!TIFFGetField ( tiff, TIFFTAG_FILLORDER, &tiff_fillorder))     ||	/* Expect FILLORDER_MSB2LSB */
			   (!TIFFGetField ( tiff, TIFFTAG_ORIENTATION, &tiff_orientation)) ){	/* Expect ORIENTATION_TOPLEFT */
			feprintf (ERROR_RED": could not read all the required tiff tags from file %s: %s\n", tifffile, strerror(errno));  /* check we have an appropriate tiff */
		}else if ( (tiff_cols!=NUM_COLS*2) || (tiff_rows!=NUM_ROWS*2) || (tiff_bps!=(8*sizeof(sample_t))) || (tiff_spp!=1) || (tiff_fillorder!=FILLORDER_MSB2LSB) || (tiff_orientation!=ORIENTATION_TOPLEFT) ) {
			feprintf (ERROR_RED": tiff not in the correct format. Tags are (name,expected,actual): (IMAGE_WIDTH,%d,%d), (IMAGE_LENGTH,%d,%d), (BITS_PER_SAMPLE,%d,%d), (SAMPLES_PER_PIXEL,%d,%d), (TIFFTAG_FILLORDER,%d,%d), (TIFFTAG_ORIENTATION,%d,%d).\n",  NUM_COLS*2,tiff_cols, NUM_ROWS*2,tiff_rows, (8*(int)sizeof(sample_t)),tiff_bps, 1,tiff_spp, FILLORDER_MSB2LSB,tiff_fillorder, ORIENTATION_TOPLEFT,tiff_orientation);
		}
		for (r = 0 ; r < (NUM_ROWS*2); r++){
			ret = TIFFReadEncodedStrip(tiff,  r,  &pixels_tiff[r],  NUM_COLS*2* sizeof(sample_t));	/* Read row r out of tiff, put into the row of pixels_tiff. NB not quadrants here. */
			if (ret < 0){
				feprintf (ERROR_RED": failed to read row %d from tiff: %s\n", r, strerror(errno));
			}
		} /* pixels_tiff[r][c] is now populated, in the X11 ordering, origin top left, (not separate quadrants, still in MSBytes) */
	}

	if (adcarg == NULL){			/* ADC mode. Default to 'ideal' */
		adcmode = ADC_MODE_IDEAL;
	}else if (!strcmp (adcarg, "ideal")){
		adcmode = ADC_MODE_IDEAL;
	}else if (!strcmp (adcarg, "noise")){
		adcmode = ADC_MODE_NOISE;
	}else if (!strcmp (adcarg, "hum")){
		adcmode = ADC_MODE_HUM;
	}else{
		feprintf(ERROR_RED": invalid adc_mode: %s. (Use: ideal|noise|hum)\n", adcarg);
	}

	eprintf ("Simulating "MAGENTA"Hawaii"NORM" chip, replaying from input file %s to output file %s....\n", infile, outfile);
	deprintf("Configuration: Reset level: %d, Reset time-constant: %.2f us, Glow: %.2f e/us, Background: %.2fus, Light: %.2f e/us, Tiff gain: %.2f e/us, Hawaii noise: %.2f e, Adc noise: %.2f e, Hum level: %.2f e, Hum frequency: %.2f Hz.\n",
		reset_level, (double)reset_tconst_ns/NS_US, glow_ens*NS_US, background_ens*NS_US, light_ens*NS_US, tiff_ens*NS_US, hawaii_noise, adc_noise, mains_hum_level, mains_fhz);
	/* End processing options */

	/* Initialise chip, set registers to garbage */
	initialise_chip();

	if (test){
		/* Create a .cam format file: reset, read, in a raster. */
		gen_cam();

	}else{
		/* Iterate over the input file, linewise. The format is documented in /usr/local/share/doc/pb_parse/pbsim.txt */
		line_num = 1; 		/* one-based (for humans) */
		while (fgets(buffer, BUFFER_LEN, input_fh) != NULL){
			ret = parse_sourceline (buffer, line_num);	/* Parse the line */
			if (ret == -1){					/* comment or blank. */
				line_num++;
				continue;
			}else if (ret != 0){				/* parse error */
				fatal_error2();
				break;
			}
			/* If line contains data, it's now in: inbus, length_ns */
			deprintf("\nSimulation step %d: input_bus: 0x%04x, length_ns: %04lld (%.2f ms)\n", instr_count, inbus, length_ns, ((float)length_ns/NS_MS));
			simulate_step(); 	/* Actually Do something with the input. */
			line_num++;
		}
	}

	/* End. Clean up */
	if (input_fh != NULL){ fclose(input_fh);}
	if (output_fh != NULL){ fclose(output_fh);}
	if (tiff != NULL){ TIFFClose(tiff); }

	/* Summarise */
	clock_gettime( CLOCK_REALTIME, &stop_time);
	run_time = (double)( stop_time.tv_sec - start_time.tv_sec ) + (double)( stop_time.tv_nsec - start_time.tv_nsec )/NS_S;
	if (conterr_count > 0){
		eprintf("\nFinished, but ignored %d errors (and %d supressed warnings). Simulated %d instructions (%.3f ms virtual), including %d read(s) and %d reset(s). Runtime: %4.2f s.\n", conterr_count, sqw_count, instr_count, (double)elapsed_ns/NS_MS, read_count, reset_count, run_time);
		return (EXIT_FAILURE);
	}else if (sqw_count){
		eprintf("\nFinished successfully (but squelched %d warnings, clipped %d times). Simulated %d instructions (%.3f ms virtual), including %d read(s) and %d reset(s). Runtime: %4.2f s.\n", sqw_count, clip_count, instr_count, (double)elapsed_ns/NS_MS, read_count, reset_count, run_time);
		return (EXIT_SUCCESS);
	}else{
		eprintf("\nFinished successfully. Simulated %d instructions (%.3f ms virtual), including %d read(s) and %d reset(s). Runtime: %4.2f s.\n", instr_count, (double)elapsed_ns/NS_MS, read_count, reset_count, run_time);
		return (EXIT_SUCCESS);
	}
}
